﻿MedwebCloud = window.MedwebCloud || {}
if (!('meta' in MedwebCloud)) {
    MedwebCloud.meta = {}
}
MedwebCloud.meta['MedwebCloud.model.GlassSetup'] = function (config, type) {
    return config;
}

/**
 * This class extends the normal Json writer so that related objects are sent down to the server on the CRUD operations.
 * The default Sencha writer only sends the data from the top level object being saved. I found two ways to send
 * down the relationship/nested data, one is to define a field using the same name as the relationship (don't do
 * this), and two to write a writer that does it (this class).
 *
 * The problem with the first solution is this.. If you define an object like this:
 *         fields: [
 *          .. other fields
 *          {name: 'anotherField', persist: false},
 *          {name: 'categories'}
 *         ],
 *         hasMany: [
 *          {model: 'helium.model.Category', name: 'categories'}
 *         ],
 *
 * In the definition above there is a field that is the same name as the relationship (hasMany). The problem is
 * that the attribute 'persist' will not be respected for any of the Category objects.. So if you want to exclude
 * attributes from being sent to the server (persisted), the above workaround won't do it.
 *
 * This class does solve that problem and you should not define a 'field' with the same name as an association.
 *
 * This is an amalgam of the ideas @:
 * http://www.sencha.com/forum/showthread.php?141957-Saving-objects-that-are-linked-hasMany-relation-with-a-single-Store
 */
Ext.define('MedwebCloud.IncludeRelationshipJsonWriter', {
    extend: 'Ext.data.writer.Json',
    /*
     * This function overrides the default implementation of json writer. Any hasMany relationships will be submitted
     * as nested objects. When preparing the data, only children which have been newly created, modified or marked for
     * deletion will be added. To do this, a depth first bottom -> up recursive technique was used.
     */
    getRecordData: function (record) {
        return record.raw;

        //Setup variables
        var isPhantom = record.phantom === true,
            writeAll = true,//this.writeAllFields || isPhantom,
            me = this, i, association, childStore, data = {};
        if (writeAll) {
            data = me.callParent(arguments);
        } else {
            var changes, name, field, fields = record.fields, nameProperty = me.nameProperty, key;
            changes = record.getChanges();
            for (key in changes) {
                if (changes.hasOwnProperty(key)) {
                    field = fields.get(key);
                    name = field[nameProperty] || field.name;
                    data[name] = changes[key];
                }
            }
            if (!record.phantom) {
                // always include the id for non phantoms
                data[record.idProperty] = record.getId();
            }
        }

        //Iterate over all the hasMany associations
        for (i = 0; i < record.associations.length; i++) {
            association = record.associations.get(i);
            if (association.get('type') != "hasmany") {
                continue;
            }
            data[association.get('name')] = [];
            childStore = record[association.get('storeName')];

            if (childStore) {
                //Iterate over all the children in the current association
                childStore.each(function (childRecord) {
                    //Recursively get the record data for children (depth first)
                    var childData = this.getRecordData.call(this, childRecord);

                    /*
                     * If the child was marked dirty or phantom it must be added. If there was data returned that was neither
                     * dirty or phantom, this means that the depth first recursion has detected that it has a child which is
                     * either dirty or phantom. For this child to be put into the prepared data, it's parents must be in place whether
                     * they were modified or not.
                     */
                    if (childRecord.dirty | childRecord.phantom | (childData != null)) {
                        data[association.get('name')].push(childData);
                        record.setDirty();
                    }
                }, me);

                /*
                 * Iterate over all the removed records and add them to the preparedData. Set a flag on them to show that
                 * they are to be deleted
                 */
                Ext.each(childStore.removed, function (removedChildRecord) {
                    //Set a flag here to identify removed records
                    removedChildRecord.set('forDeletion', true);
                    var removedChildData = this.getRecordData.call(this, removedChildRecord);
                    data[association.get('name')].push(removedChildData);
                    record.setDirty();
                }, me);
            }
        }
        //Only return data if it was dirty, new or marked for deletion.
        if (record.dirty | record.phantom | record.get('forDeletion')) {
            return data;
        }
    }
});